package clientutil

import (
	"errors"
	"net/http"
	"time"

	"github.com/cenkalti/backoff"
)

// NewExponentialBackOff creates a backoff.ExponentialBackOff object
// with our own default values.
func NewExponentialBackOff() *backoff.ExponentialBackOff {
	b := backoff.NewExponentialBackOff()
	b.InitialInterval = 100 * time.Millisecond
	//b.Multiplier = 1.4142
	return b
}

// A temporary (retriable) error is something that has a Temporary method.
type tempError interface {
	Temporary() bool
}

type tempErrorWrapper struct {
	error
}

func (t tempErrorWrapper) Temporary() bool { return true }

// TempError makes a temporary (retriable) error out of a normal error.
func TempError(err error) error {
	return tempErrorWrapper{err}
}

// Retry operation op until it succeeds according to the backoff
// policy b.
//
// Note that this function reverses the error semantics of
// backoff.Operation: all errors are permanent unless explicitly
// marked as temporary (i.e. they have a Temporary() method that
// returns true). This is to better align with the errors returned by
// the net package.
func Retry(op backoff.Operation, b backoff.BackOff) error {
	innerOp := func() error {
		err := op()
		if err == nil {
			return err
		}
		if tmpErr, ok := err.(tempError); ok && tmpErr.Temporary() {
			return err
		}
		return backoff.Permanent(err)
	}
	return backoff.Retry(innerOp, b)
}

var errHTTPBackOff = TempError(errors.New("temporary http error"))

func isStatusTemporary(code int) bool {
	switch code {
	case http.StatusTooManyRequests, http.StatusBadGateway, http.StatusServiceUnavailable, http.StatusGatewayTimeout:
		return true
	default:
		return false
	}
}

// RetryHTTPDo retries an HTTP request until it succeeds, according to
// the backoff policy b. It will retry on temporary network errors and
// upon receiving specific temporary HTTP errors. It will use the
// context associated with the HTTP request object.
func RetryHTTPDo(client *http.Client, req *http.Request, b backoff.BackOff) (*http.Response, error) {
	var resp *http.Response
	op := func() error {
		// Clear up previous response if set.
		if resp != nil {
			resp.Body.Close()
		}

		var err error
		resp, err = client.Do(req)
		if err == nil && isStatusTemporary(resp.StatusCode) {
			resp.Body.Close()
			return errHTTPBackOff
		}
		return err
	}

	err := Retry(op, backoff.WithContext(b, req.Context()))
	return resp, err
}
